@@ -1,6 +1,11 @@
 from collections import defaultdict
-import re
+import itertools
 import sys
+from bs4.element import (
+    CharsetMetaAttributeValue,
+    ContentMetaAttributeValue,
+    whitespace_re
+    )
 
 __all__ = [
     'HTMLTreeBuilder',
@@ -137,6 +142,24 @@ class TreeBuilder(object):
     def set_up_substitutions(self, tag):
         return False
 
+    def _replace_cdata_list_attribute_values(self, tag_name, attrs):
+        """Replaces class="foo bar" with class=["foo", "bar"]
+
+        Modifies its input in place.
+        """
+        if self.cdata_list_attributes:
+            universal = self.cdata_list_attributes.get('*', [])
+            tag_specific = self.cdata_list_attributes.get(
+                tag_name.lower(), [])
+            for cdata_list_attr in itertools.chain(universal, tag_specific):
+                if cdata_list_attr in dict(attrs):
+                    # Basically, we have a "class" attribute whose
+                    # value is a whitespace-separated list of CSS
+                    # classes. Split it into a list.
+                    value = attrs[cdata_list_attr]
+                    values = whitespace_re.split(value)
+                    attrs[cdata_list_attr] = values
+        return attrs
 
 class SAXTreeBuilder(TreeBuilder):
     """A Beautiful Soup treebuilder that listens for SAX events."""
@@ -218,9 +241,6 @@ class HTMLTreeBuilder(TreeBuilder):
         "output" : ["for"],
         }
 
-    # Used by set_up_substitutions to detect the charset in a META tag
-    CHARSET_RE = re.compile("((^|;)\s*charset=)([^;]*)", re.M)
-
     def set_up_substitutions(self, tag):
         # We are only interested in <meta> tags
         if tag.name != 'meta':
@@ -235,27 +255,22 @@ class HTMLTreeBuilder(TreeBuilder):
         # tags that provide the "charset" attribute. It also means
         # HTML 4-style <meta> tags that provide the "content"
         # attribute and have "http-equiv" set to "content-type".
+        #
+        # In both cases we will replace the value of the appropriate
+        # attribute with a standin object that can take on any
+        # encoding.
         meta_encoding = None
         if charset is not None:
             # HTML 5 style:
             # <meta charset="utf8">
             meta_encoding = charset
-
-            # Modify the tag.
-            tag['charset'] = "%SOUP-ENCODING%"
+            tag['charset'] = CharsetMetaAttributeValue(charset)
 
         elif (content is not None and http_equiv is not None
               and http_equiv.lower() == 'content-type'):
             # HTML 4 style:
             # <meta http-equiv="content-type" content="text/html; charset=utf8">
-            match = self.CHARSET_RE.search(content)
-            if match is not None:
-                meta_encoding = match.group(3)
-
-                # Modify the tag.
-                def rewrite(match):
-                    return match.group(1) + "%SOUP-ENCODING%"
-                tag['content'] = self.CHARSET_RE.sub(rewrite, content)
+            tag['content'] = ContentMetaAttributeValue(content)
 
         return (meta_encoding is not None)
 
@@ -276,7 +291,7 @@ def register_treebuilders_from(module):
 # builder registrations will take precedence. In general, we want lxml
 # to take precedence over html5lib, because it's faster. And we only
 # want to use HTMLParser as a last result.
-from .import _htmlparser
+from . import _htmlparser
 register_treebuilders_from(_htmlparser)
 try:
     from . import _html5lib
